// N6502 - Nano 6502 for neural nets. Target about 250 LUTS/core + 1Kbyte
// 
// (C) 2017 Microingress Ltd
// renato.bugge<remove>@microingress.eu
// All Rights Reserved.
//
//	Mini6502.v
//
// This source file is free software: you can redistribute it and/or modify 
// it under the terms of the GNU Lesser General Public License as published 
// by the Free Software Foundation, either version 3 of the License, or     
// (at your option) any later version.                                      
//                                                                          
// This source file is distributed in the hope that it will be useful,      
// but WITHOUT ANY WARRANTY; without even the implied warranty of           
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            
// GNU General Public License for more details.                             
//                                                                          
// You should have received a copy of the GNU General Public License        
// along with this program.  If not, see <http://www.gnu.org/licenses/>.    
//                                                                          

// Memory implementation is 1KByte/core:
// $000-$1FF Code memory of 256 word (16-bit). Each opcode takes 2 bytes.
// $200-$2FF Data memory
// $300-$3FF Stack memory

// Memory is dual-port: Read happends at $000-$1FF, Write at $200-$3FF. No self-code modifications.
// Two alternate ways:
// 1) Each 1K memory is one core only. Dual port with read/write at same cycle. All memory will be 8-bits so two
//    memory cycles per instruction. Alternate 16-bit can enable one memory cycle for Code memory and data reading,
//    but for data write (8-bit), one will need to read the 16-bit word first, then modify it and write it back.
// 2) Each 1K memory is shared for two cores. First half for 1st core, second for 2nd core. 16-bit code memory read
//    happends in one 1KB block, 8-bit data/stack memory read/write happends in another 1KB block. Dual port memory
//    guarantees access for both cores without higher frequency access.
// In both cases, the 6502 controller needs to access both code memory and data memory. For code memory we would
// stop the N6502 core before the code is changed, which makes read/write cycles available. For data/stack memory,
// the N6502 would probably be running, but not all instructions are write/read which allows the controller access
// at a lower rate. Stack memory usually gives more access time and less congestion.
// 3) An alternate to 2 is to only allow one core per 1K memory, but have an additional 1K data, so that each core 
//    has 2KB. With dual port, this allows full access to the 6502 controller without interruption.

// Since a N6502 has 8-bit addressing and 9-bit code memory, assembly branches will need modification to work 
// correctly. This can be implemented in software as the 6502 controller must copy the code to each N6502 code// memory (since the N6502 in fact can't access its own code memory). E.g a BCC *-7 will branch back 7 WORDS and not
// 7 bytes.
// Since all opcodes are 2-bytes, PHA/PLA opcodes require an extra byte after the opcode. If the extra byte is $ff,
// this accesses the stack of the current N6502. Other numbers are reserved for future multicore implementation.

// Logic works this way after start:
// 1) Read two bytes at PC (opcode+operand) in code mem - 1 cycle
// 2) Decode opcode in cpu --> Write/Read, modify PC - 1 cycle
// 3) Write/Read data in data memory & Read two bytes at PC (opcode+operand) in code mem (Dual port mem)
// 4) Modify register (accumulator, x-register) from prev opcode & Decode opcode in cpu --> Write/Read, modify PC
// repeat 3) and 4)
// Since code and data mem can be accessed in same cycle, decoding of opcode can be done while next code mem bytes
// are read. It also means that accumulator & x-register modifications from direct data should wait until next cycle
// (since ZP data will be available in next cycle.

// If accumulator, x-register and datamemory is increased to 16-bit, most things will be done as 16 bit. Only operand
// remains at 8-bit --> No Immediate instructions allowed for 16-bit (or alternately we can add a LDB# which is MSB.

// For neural nets, using 16-bit is very helpful to reduce code size and improve accuracy.
// Wanted for neural nets: MUL and DIV. 8-bit MUL requires 64KB array to be 1-cycle. DIV is more complicated.
// Due to the size of these instructions, they will have to be implemented differently.

`timescale 1ns / 100ps
module c65n02(address,datain1,datain2,addressout,dataout,we,rwdatain,rst,clk);

input [7:0] datain1;
input [7:0] datain2;
input [7:0] rwdatain;
output [7:0] dataout;
output [7:0] address;
output reg [8:0] addressout;
output we;
input rst;
input clk;

reg [8:0] accumulator; // keep carry in accumulator bit 8
reg [7:0] pc;
reg [8:0] opcode;  // keep carry in opcode bit 8 (for correct case determination)
reg [7:0] operand;
reg [7:0] xregister;
wire [7:0] opdatain;
reg we_cpu;
reg [7:0] branch; //branch "flag"
reg [7:0] stackpointer;
reg [15:0] mulregister;
reg [15:0] sqrmem [0:255];
reg [7:0] i;

initial begin
//	$display("time, state, adreg, pc, rst, clk, data, adress");
//	$monitor($stime,",pc=", pc, ",rst=", rst,",clk=",clk,",datain1=",datain1,",datain2=",datain2,",dataout=",dataout,",address=",address); 
		$readmemh("sqrdata.txt", sqrmem);
//		for (i=0; i < 256; i=i+1) $display("%d:%h",i,sqrmem[i]);
end

// Note: #Immediate for opcode[2]=1'b0
//		 $ZP/$ZP,X for opcode[2]=1'b1
// So select between indata (rwdatain) for ZP response or immediate value (datain2)
assign opdatain=opcode[2]?rwdatain:datain2;

always @(negedge clk)
		if (rst==1'b1) begin
			accumulator <= 0;	
			pc <= 8'hff;
//			opcode <= 0;
			operand <= 0;
			xregister <= 0;
			branch <= 8'h00;
		end
		else begin
			// ALU 
			//
			// LDA #nn		is $A9, %10101001
			// LDA $ZP 		is $A5, %10100101
			// LDA $ZP,X	is $B5, %10110101
			// LDX #nn		is $A2, %10100010
			// LDX $ZP		is $A6, %10100110
			// STA $ZP		is $85, %10000101
			// STA $ZP,X	is $95, %10010101
			// STX $ZP		is $86, %10000110
			// BCC $xx		is $90, %10010000
			// BCS $xx		is $B0, %10110000
			// ADC #nn		is $69, %01101001
			// ADC $ZP		is $65, %01100101
			// ADC $ZP,X	is $75, %01110101
			// SBC #nn		is $E9, %11101001  			
			// SBC $ZP		is $E5, %11100101
			// SBC $ZP,X	is $F5, %11110101   
			// CMP #nn		is $C9, %11001001
			// CMP $ZP		is $C5, %11000101
			// CPX #nn		is $E0, %11100000
			// CPX $ZP		is $E4, %11100100
			// DEX			is $CA, %11001010
			// AND #nn		is $29, %00101001
			// AND $ZP		is $25, %00100101
			// AND $ZP,X	is $35, %00110101
			// ORA #nn		is $09, %00001001
			// ORA $ZP		is $05, %00000101
			// ORA $ZP,X	is $15, %00010101
			// CLC			is $18, %00011000
			// SEC			is $38, %00111000
			// TXA			is $8A, %10001010
			// TAX			is $AA, %10101010
			// ROL A		is $2A, %00101010
			// ROR A		is $6A, %01101010
			// ASL A        is $0A, %00001010
			// LSR A        is $4A, %01001010
			// PHA			is $48, %01001000
			// PLA 			is $68, %01101000
			// SQR #n		is $49, %01001001
			// SQR $ZP		is $45, %01000101
			// SQR $ZP,X	is,$55, %01010101
			
			// Note: $AD, $BD, $CD, $6D, $7D, $3D, $2D, $1D and $0D also triggers (as $x5 instructions)
			// NOP : all other instructions
			//accumulator[7:0] <= operand;
			pc <= pc + 1'b1; 
			casez(opcode[7:0])  // Do last cycle opcode (as we now have $ZP address content)
			// LDA #nn is $A9, %10101001
			// LDA $ZP is $A5, LDA $ZP,X is $B5
				8'b101???01 : accumulator[7:0] <= opdatain;
			// LDX #nn		is $A2, %10100010
			// LDX $ZP		is $A6, %10100110
				8'b10100?10 : xregister <= opdatain;
			// STA $ZP		is $85, %10000101
			// STA $ZP,X	is $95, %10010101
				8'b100?0101 : operand<=accumulator[7:0];					// STA $ZP or $ZP,X
			// STX $ZP		is $86, %10000110
				8'b10000110 : operand<=xregister;							// STX $ZP
			// ADC #n: $69
				8'b011???01 : accumulator <= {1'b0, accumulator[7:0]} + {1'b0, opdatain};
			// ADC $ZP: $65, $ZP,X: $75
//				9'b?011?0101 : accumulator <= {1'b0, accumulator[7:0]} + {1'b0, rwdatain};
			// SBC #nn		is $E9, %11101001  
				8'b111???01 : accumulator <= accumulator[8:0] - {1'b0, opdatain};
			// SBC $ZP		is $E5, %11100101,$ZP	is $F5, %11110101   
//				9'b?111?0101 : accumulator <= accumulator[8:0] - {1'b0, rwdatain};
			// CMP #nn		is $C9, %11001001
			// CMP $ZP		is $C5, %11000101
				8'b1100??01 : accumulator[8] <= (accumulator[7:0]>=opdatain);
			// CPX #nn		is $E0, %11100000
			// CPX $ZP		is $E4, %11100100
				8'b11100?00 : accumulator[8] <= (xregister>=opdatain);			
			// DEX			is $CA, %11001010
				8'b11001010 : xregister <= xregister - 1;					
			// AND #nn		is $29, %00101001
				8'b001???01 : accumulator <= {accumulator[8],(accumulator[7:0]&opdatain)};
			// AND $ZP		is $25, %00100101, AND $ZP,X	is $35, %00110101
//				9'b?001?0101 : accumulator <= {accumulator[8],(accumulator[7:0]&rwdatain)};
			// ORA #n: $09
				8'b000???01 : accumulator <= {accumulator[8],(accumulator[7:0]|opdatain)};
			// ORA $ZP: $05, $ZP,X: $15
//				9'b?000?0101 : accumulator <= {accumulator[8],(accumulator[7:0]|rwdatain)};
			// CLC			is $18, %00011000
			// SEC			is $38, %00111000
				8'b00?11000 : accumulator[8] <= opcode[5]; // clear/set carry
//				9'b000111000 : accumulator[8] <= 1'b1; // set carry
			// TXA			is $8A, %10001010
				8'b10001010 : accumulator <= {accumulator[8],xregister}; // X to A
			// TAX			is $AA, %10101010
				8'b10101010 : xregister <= accumulator[7:0]; // A to X
			// TSX			is $BA, %10111010
				8'b10111010 : xregister <= stackpointer; // S to X			
			// TXS			is $9A, %10011010
				8'b10011010 : stackpointer <= xregister; // X to S	
			// ROL #n		is $2A, %00101010
			// ASL A        is $0A, %00001010
				8'b00?01010 : accumulator <= {accumulator[7:0],(accumulator[8]&opcode[5])}; // ROL/ASL
			// ROR #n		is $6A, %01101010
			// LSR A        is $4A, %01001010
				8'b01?01010 : accumulator <= {accumulator[0],(accumulator[8]&opcode[5]),accumulator[7:1]}; // ROR/LSR
			// PHA			is $48, %01001000
				8'b01001000 : begin operand<=accumulator[7:0]; stackpointer<=stackpointer+1; end  // Write accumulator to stack
			// PLA 			is $68, %01101000
				8'b01101000 : begin accumulator[7:0]<=opdatain; stackpointer<=stackpointer-1; end
				default: ; //NOP 
			endcase
/*			casez(opcode)
			// BCC $REL		is $90
				9'b010010000 : begin pc <= pc + operand; branch<=8'b11111111; end // BCC branch taken (carry is clear)
			// BCS $REL		is $B0
				9'b110110000 : begin pc <= pc + operand; branch<=8'b11111111; end // BCS branch taken (carry is set)
				default : branch <=8'b00000000;
			endcase*/
			if ((opcode==9'b010010000) | (opcode==9'b110110000))
				begin pc <= pc + {opdatain[7],opdatain[7:1]}; branch<=8'b11111111; end
			else
				begin branch <=8'b00000000; end
			// SQR #n		is $49, %01001001
			// SQR $ZP		is $45, %01000101
			// SQR $ZP,X	is,$55, %01010101
			// 8'b010???01
			if (mulregister!=0)
			begin accumulator<={accumulator[8],mulregister[15:8]};
				xregister<=mulregister[7:0]; end
			if((opcode[7:5]==3'b010) && (opcode[1:0]==2'b01)) 
//				mulregister <= (accumulator[7:0]*opdatain);
				mulregister <= sqrmem[accumulator[7:0]];
			else mulregister <= 16'h0000;

//			Going from 9-bit to 6-bit case statement will reduce Lattice compilation with 1 LUT per case, but requires
//			some opcode translating through an array (so probably slower and larger for small number of cores)

/*			case({accumulator[8], datain1[5:4],datain1[2:0]})
				6'b010001 : begin pc <= pc + 1'b1; accumulator 	 <= {1'b0, accumulator[7:0]} + {1'b0, datain2}; end // add
				6'b000011 : begin pc <= pc + 1'b1; accumulator[7:0] <= ~(accumulator[7:0]|datain2); end // nor
				6'b101000 : begin pc <= pc + 1'b1; accumulator[8]   <= 1'b0; end// branch not taken, clear carry					   
				6'b001000 : pc <= datain2[7:0]; // BCC branch taken (carry is clear)
				default: pc <= pc + 1'b1; 
			endcase							// default:  instruction fetch, bcc taken
*/
			end
	
always @(posedge clk)
	if (rst==1'b1) begin
		opcode <= 0;
		addressout<= 0;
	//	stackpointer<= 0;
	end
	else 
	begin
		// Will always read/write address at addressout next (even for #XX which is not used)
		// The read/write is in the last phase of a CPU cycle
		// STA $ZP		is $85, %10000101
		// STA $ZP,X	is $95, %10010101
		if (datain1[4:0]==5'h95)	
	//	if (datain1==8'h95)
			addressout<={1'b0,(datain2[7:0]+xregister[7:0])};  // $ZP,X (also STA)
		else if (datain1[4:0]==5'b01000)
			addressout<={1'b1,stackpointer+opcode[6]};				// PHA or PLA (opcode[6]=1 for PLA)
		else addressout<={1'b0,datain2}; 					// $ZP or #NN  


/*
		casez (opcode[5:0])
			6'b?10101: addressout<={1'b0,(datain2[7:0]+xregister[7:0])};  // STA $ZP,X
			6'b001000: begin addressout<={1'b1,stackpointer}; stackpointer=stackpointer+1; end	//PHA
			6'b101000: begin addressout<={1'b1,stackpointer-1}; stackpointer=stackpointer-1; end	//PHA
			default: addressout<={1'b0,datain2};
		endcase
*/		

		
		// Note that branch is one cycle late, so next instruction (in datain now), will need to be
		// NOP'ed, but only if branch has been taken (e.g. branch=8'b11111111):
		opcode<={accumulator[8], datain1[7:0]|branch};
		/*		case (branch)
			1'b0: begin opcode<={accumulator[8], datain1[7:0]}; end
			1'b1: begin opcode<={accumulator[8], 8'b11111111}; end
		endcase*/
		
		// Need to set Write Enable early to catch write
		// STA $ZP		is $85, %10000101
		// STA $ZP,X	is $95, %10010101
		// STX ZP,Y		is $96, 8'b10010110:  (ikke i bruk)
		// STX ZP		is $86, %10000110:
		// PHA			is $48, %01001000:
		// PLA 			is $68, %01101000		
		casez(datain1)
		//	6'b100?01: we_cpu=1'b1;					// STA $ZP or $ZP,X or STX $ZP
			8'b100?01??: we_cpu=1'b1;					// STA $ZP or $ZP,X or STX $ZP

		//	8'b10000110: we_cpu=1'b1;							// STX $ZP
			8'b01001000: we_cpu=1'b1;				// PHA
			default: we_cpu=1'b0;
		endcase

	end
// output
assign address = pc;
assign dataout = operand; 	// Dataout is always the operand (which is written for STA/STX)
assign we = we_cpu;

endmodule
